from abc import abstractmethod


class Messenger:

    def __init__(self, step_by_message_count=100):
        self.step_by_message_count = step_by_message_count
        pass

    @abstractmethod
    def send_message(self, subject: str, body: str):
        pass


class _DefaultMessenger(Messenger):
    def __init__(self, step_by_message_count=100):
        super().__init__(step_by_message_count)

    def send_message(self, subject: str, body: str):
        print(subject, body)
        pass



import torch
from torch import Tensor
import torch.nn as nn
from .PositionalEncoding import PositionalEncoding

from .progress import LearningProgress


class MORTM(nn.Module):
    token_dict = {
        0: "600_633",
        1: "10_139",
        2: "300_429",
        3: "500_600",
        4: "600_605"
    }

    def __init__(self, vocab_size, progress: LearningProgress, trans_layer=6, num_heads=8, d_model=512,
                 dim_feedforward=1024, dropout=0.1,
                 position_length=2048):
        super(MORTM, self).__init__()

        self.progress = progress
        self.trans_layer = trans_layer
        self.num_heads = num_heads
        self.d_model = d_model
        self.dim_feedforward = dim_feedforward
        self.dropout = dropout

        #位置エンコーディングを作成
        self.positional: PositionalEncoding = PositionalEncoding(self.d_model, dropout, position_length).to(
            self.progress.get_device())
        #Transformerの設定
        self.transformer: nn.Transformer = nn.Transformer(d_model=self.d_model, nhead=num_heads,  #各種パラメーターの設計
                                                          num_encoder_layers=self.trans_layer,
                                                          num_decoder_layers=self.trans_layer,
                                                          dropout=self.dropout, dim_feedforward=dim_feedforward,
                                                          ).to(self.progress.get_device())
        print(f"Input Vocab Size:{vocab_size}")
        self.Wout: nn.Linear = nn.Linear(self.d_model, vocab_size).to(self.progress.get_device())

        self.embedding: nn.Embedding = nn.Embedding(vocab_size, self.d_model).to(self.progress.get_device())
        self.softmax: nn.Softmax = nn.Softmax(dim=-1).to(self.progress.get_device())

    def forward(self, inputs_seq, tgt_seq, input_mask, tgt_mask, input_padding_mask, tgt_padding_mask):

        inputs_em: Tensor = self.embedding(inputs_seq)
        inputs_em = inputs_em.permute(1, 0, 2)
        inputs_pos: Tensor = self.positional(inputs_em)

        if tgt_seq is None:
            tgt_pos = inputs_pos
        else:
            tgt_em: Tensor = self.embedding(tgt_seq)
            tgt_em = tgt_em.permute(1, 0, 2)
            tgt_pos: Tensor = self.positional(tgt_em)

        #print(inputs_pos.shape, tgt_pos.shape)
        out: Tensor = self.transformer(inputs_pos, tgt_pos, input_mask, tgt_mask,
                                       src_key_padding_mask=input_padding_mask, tgt_key_padding_mask=tgt_padding_mask)

        out.permute(1, 0, 2)

        score:Tensor = self.Wout(out)
        return score.to(self.progress.get_device())

    def generate_by_length(self, input_seq, max_length, p=0.9, temperature=0.1):
        self.eval()
        output = torch.tensor([input_seq], dtype=torch.long).unsqueeze(1).to(self.progress.get_device())
        for _ in range(max_length):
            with torch.no_grad():
                output = self._next_note_token(output)

        return output

    def _next_note_token(self, output, p=0.9, temperature=0.1):
        isEnd = False
        token_count = 0
        while not isEnd:
            mask = self.transformer.generate_square_subsequent_mask(output.shape[1]).to(self.progress.get_device())
            outputs = self(output, output, mask, mask, None, None)
            logits = outputs[:, -1, :]

            str_token_duration = self.token_dict[token_count]
            parts = str_token_duration.split("_")
            token_duration = [int(part) for part in parts]
            if token_count != 4:
                token = self._next_token(logits[:, token_duration[0]:token_duration[1]])
                token += token_duration[0]
            else:
                token = self._next_token(logits)
                isEnd = True
            token_count += 1

            output = torch.cat((output.flatten(), token.unsqueeze(0))).unsqueeze(0).to(self.progress.get_device())

        return output

    def _next_token(self, end_logit_score: Tensor, p=0.9, temperature=0.1):
        end_logit_score = end_logit_score / temperature

        sorted_logits, sorted_indices = torch.sort(end_logit_score, descending=True)
        cumulative_probs = torch.cumsum(self.softmax(sorted_logits), dim=-1)
        sorted_indices_to_remove = cumulative_probs > p
        sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
        sorted_indices_to_remove[..., 0] = 0
        indices_to_remove = sorted_indices[sorted_indices_to_remove]
        end_logit_score[:, indices_to_remove] = -float('Inf')

        score = self.softmax(end_logit_score)[-1]
        dis = torch.distributions.categorical.Categorical(probs=score)
        next_token = dis.sample()

        #print(next_token)
        return next_token

    def top_p_sampling(self, input_ids, tokenizer, p=0.9, max_length=20, temperature=0.2):
        self.eval()
        output = torch.tensor([input_ids], dtype=torch.long).unsqueeze(1).to(self.progress.get_device())
        for _ in range(max_length):
            with torch.no_grad():
                mask = self.transformer.generate_square_subsequent_mask(output.shape[1]).to(self.progress.get_device())
                outputs = self(output, output, mask, mask, None, None)
                logits = outputs[:, -1, :]

                #print(logits[-1, 10:138])

                logits = logits / temperature

                sorted_logits, sorted_indices = torch.sort(logits, descending=True)
                cumulative_probs = torch.cumsum(self.softmax(sorted_logits), dim=-1)
                sorted_indices_to_remove = cumulative_probs > p
                sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
                sorted_indices_to_remove[..., 0] = 0
                indices_to_remove = sorted_indices[sorted_indices_to_remove]
                logits[:, indices_to_remove] = -float('Inf')

                probabilities = self.softmax(logits)[-1]
                dis = torch.distributions.categorical.Categorical(probs=probabilities)
                next_token = dis.sample()
                # バッチサイズを一致させるために次元を調整
                next_token = next_token.unsqueeze(0)
                output = torch.cat((output.flatten(), next_token)).unsqueeze(0).to(self.progress.get_device())
                #print(output)
        return output.tolist()


class DummyDecoder(nn.Module):
    def __init__(self):
        super(DummyDecoder, self).__init__()

    def forward(self, tgt, memory, tgt_mask, memory_mask, tgt_key_padding_mask, memory_key_padding_mask, **kwargs):
        return memory



'''
Midiデータを tokenizerで変換し、数値の配列にしたものをnumpyに変換し、指定されたディレクトリに保存します。
'''

from typing import Any

import mido.midifiles.meta
from numpy import ndarray, dtype
from pretty_midi import PrettyMIDI, Instrument, Note
import numpy as np

from . import constants
from AGSM.convert import ConvTempo, ConvKey
from .tokenizer import Tokenizer
from . import tokenizer as tr

PITCH = 0
VELOCITY = 1
DURATION_INT = 2
DURATION_FEW = 3
BEGIN_TIME_INT = 4
BEGIN_TIME_FEW = 5
START_SEQ = np.array([[0, 0, 0, 0]], dtype=int)
END_SEQ = np.array([[5, 5, 5, 5]], dtype=int)
''''
規則：
    AyaNode:
        Aya_node = [音高, 強さ, 始まりの拍(16), 終わりの拍までの長さ(16)]
    TOKEN:
        <S_SEQ>: [0, 0, 0, 0]
        <E_SEQ>: [5, 5, 5, 5]
'''

def convert_tolist(target: list):
    new_target = None
    for t in target:
        t: ndarray
        cul = []
        for c in t:
            cul.append(c)
        if new_target is not None:
            new_target = new_target + [cul]
        else:
            new_target = [cul]

    return new_target


class ConvertMidi:

    def __init__(self, tokenizer: Tokenizer, directory: str, file_name: str, program_list: list,
                 tempo: int, midi_data: PrettyMIDI = None):
        self.directory = directory
        self.file_name = file_name
        self.midi_data = midi_data
        self.program_list = program_list
        self.tempo = tempo
        self.aya_node = None
        self.is_error = False
        self.tokenizer = tokenizer

    def convert(self):
        self._ct_tempo()
        #self._ct_key()
        if not self.is_error:
            convert_notes: np = None
            program_count = 0
            for inst in self.midi_data.instruments:
                inst: Instrument = inst
                if not inst.is_drum and inst.program in self.program_list:
                    cn = self.ct_aya_node(inst.notes)

                    if len(cn) >= 1000:
                        new_array = np.array_split(cn, np.arange(1000, len(cn), 1000))
                        new_array = convert_tolist(new_array)
                        cn: ndarray = np.array(self._padding(new_array))
                        print(cn.shape)
                    else:
                        cn: ndarray = np.array(self._padding([cn.tolist()], 1000))

                    if convert_notes is None:
                        convert_notes = cn
                    else:
                        #pad_convert_notes, pad_cn = self.padding(convert_notes, cn)
                        convert_notes = np.vstack((convert_notes, cn), dtype=int)
                    program_count += 1

            if program_count == 0:
                print(f"{self.directory}/{ self.file_name}には欲しい楽曲がありませんでした")
                self.is_error = True
            elif np.min(convert_notes) < 0:
                print(f"{self.directory}/{ self.file_name}の値の中に0以下の値が格納されていたため、中断します。")
                self.is_error = True

            self.aya_node = convert_notes
            return convert_notes


    def split_seq_data(self, target: ndarray):

        if len(target.shape) == 1:
            musics_seq = [target.tolist()]
        else:
            musics_seq = target.tolist()

        new_music_seq = [[]]
        for i in range(len(musics_seq)):
            result = []
            current_sublist = []
            for value in musics_seq[i]:
                current_sublist.append(value)
                if value == 2:
                    result.append(current_sublist)
                    current_sublist = []
            new_music_seq = new_music_seq + result

        return np.array(self._padding(new_music_seq))

        pass

    def _padding(self, target: list, max_length: int = None):
        if max_length is None:
            max_lengths = []
            for t in target:
                max_lengths.append(len(t))
            max_length = max(max_lengths)

        print(f"Max length is {max_length}")
        for t in target:
            if len(t) < max_length:
                for _ in range(max_length - len(t)):
                    t.append(0)
        return target
        pass

    def save(self, save_directory: str) -> bool:
        if not self.is_error:
            print(f"Result shape is:{self.aya_node.shape}")

            np.savez(save_directory + "/" + self.file_name, self.aya_node)
            print("処理が正常に終了しました。")
            return True
        else:
            print("Transformerが望むデータ形式ではないため、保存ができませんでした。")
            return False

    def ct_aya_node(self, notes: list) -> ndarray[Any, dtype[Any]]:
        node = np.array([self.tokenizer.get(-1, constants.START_SEQ_TOKEN)])

        back_start = None
        for note in notes:

            note: Note = note
            start = self.ct_time_to_beat(note.start)
            end = self.ct_time_to_beat(note.end)

            if back_start is not None:
                shift = abs((back_start // 32) - (start // 32))
                if 4 > shift:
                    #if shift > 0:
                    node = np.append(node, self.tokenizer.get(shift, tr.SHIFT_TYPE)) #何小節のブランクができたかを計算
                else:
                    node = np.append(node, self.tokenizer.get(-1, constants.END_SEQ_TOKEN))
                    node = np.append(node, self.tokenizer.get(-1, constants.START_SEQ_TOKEN))

            pitch_token = self.tokenizer.get(note.pitch, tr.PITCH_TYPE)
            velocity_token = self.tokenizer.get(note.velocity, tr.VELOCITY_TYPE)
            duration_token = self.tokenizer.get(min(abs(end - start), 99), tr.DURATION_TYPE)
            start_token = self.tokenizer.get(start % 32, tr.START_TYPE)
            note_token = np.array([start_token, pitch_token, velocity_token, duration_token])
            node = np.concatenate((node, note_token))
            back_start = start

        node = np.append(node, self.tokenizer.get(-1, constants.END_SEQ_TOKEN))
        return node

    def _ct_key(self):
        try:
            if self.midi_data is None:
                ct = ConvKey(self.directory + "/" + self.file_name, "C")
                ct.convert()
                self.midi_data = ct.midi_data
                if ct.is_Error:
                    self.is_error = ct.is_Error
            else:
                ct = ConvKey(directory=self.directory + "/" + self.file_name, midi_data=self.midi_data, key="C")
                ct.convert()
                self.midi_data = ct.midi_data
        except OSError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except IndexError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except mido.midifiles.meta.KeySignatureError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except ValueError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True



    def _ct_tempo(self):
        try:
            if self.midi_data is None:
                ct = ConvTempo(self.directory + "/" + self.file_name, self.tempo)
                ct.convert()
                self.midi_data = ct.midi_data
                if ct.is_Error:
                    self.is_error = ct.is_Error
            else:
                ct = ConvTempo(directory=self.directory + "/" + self.file_name, midi_data=self.midi_data, change_tempo=120)
                ct.convert()
                self.midi_data = ct.midi_data
        except OSError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except IndexError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except ValueError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True

        except mido.midifiles.meta.KeySignatureError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True

        except EOFError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True

        except KeyError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True
        except ZeroDivisionError:
            print(f"{self.directory}/{ self.file_name}でエラーが発生。処理を中断します。")
            self.is_error = True



    def ct_time_to_beat(self, time: float) -> int:
        b4 = 60 / self.tempo
        b8 = b4 / 2
        b16 = b8 / 2
        b32 = b16 / 2

        beat, sub = self.calc_time_to_beat(time, b32)

        return beat

    def calc_time_to_beat(self, time, beat_time) -> (int, int):
        main_beat: int = time // beat_time
        sub_time: int = time % beat_time
        return main_beat, sub_time

    def padding(self, array1: ndarray, array2: ndarray):
        print(f"a:{array1.shape[-1]}  b:{array2.shape[-1]}")

        max_len = max(array1.shape[-1], array2.shape[-1])

        # 0でパディングして同じ長さに揃える
        if len(array1.shape) > 1:
            padded_array1 = np.array([np.pad(row, (0, max_len - len(row)), mode='constant') for row in array1])
        else:
            padded_array1 = np.pad(array1, (0, max_len - len(array1)), mode='constant')

        if len(array2.shape) > 1:
            padded_array2 = np.array([np.pad(row, (0, max_len - len(row)), mode='constant') for row in array2])
        else:
            padded_array2 = np.pad(array2, (0, max_len - len(array2)), mode='constant')

        return padded_array1, padded_array2




import torch
from torch.optim import Adam
from torch.optim.lr_scheduler import LambdaLR


def noam_lr(d_model: int, warmup_steps=4000):
    def lr_lambda(step):
        step = max(1, step)
        lr = (d_model ** -0.5) * min(step ** -0.5, step * (warmup_steps ** -1.5))
        #print(f"steps:{step}  lr:{lr}")
        return lr
    return lr_lambda



from abc import abstractmethod
import torch


class LearningProgress:

    @abstractmethod
    def step_optimizer(self, optimizer, **kwargs):
        pass

    @abstractmethod
    def get_device(self):
        pass


class _DefaultLearningProgress(LearningProgress):

    def get_device(self):
        if torch.cuda.is_available():
            return torch.device('cuda')
        else:
            return torch.device('cpu')
        pass

    def step_optimizer(self, optimizer, **kwargs):
        optimizer.step()  # オプティマイザを更新
        optimizer.zero_grad()
        pass



import json

from . import constants

'''旋律トークン'''
PITCH_TYPE = 'p'
VELOCITY_TYPE = 'v'
DURATION_TYPE = 'd'

'''指示トークン'''
START_TYPE = 's'
SHIFT_TYPE = 'h'


class Tokenizer:
    def __init__(self, save_directory: str, load_data: str = None):
        self.save_directory = save_directory
        if load_data is None:
            # 特殊トークン
            self.special_token_position = 3
            #旋律トークン
            self.pitch_position = 10
            self.velocity_position = 300
            self.duration_position = 500
            #指示トークン
            self.instruction_start_position = 600
            self.instruction_shift_position = 650

            self.tokens: dict = dict()
            self.token_max: dict = self._init_mx_dict()
            self.tokens[constants.PADDING_TOKEN] = 0
            self.tokens[constants.START_SEQ_TOKEN] = 1
            self.tokens[constants.END_SEQ_TOKEN] = 2
        else:
            with open(load_data, 'r') as file:
                self.tokens: dict = json.load(file)
                self.rev_tokens: dict = {v: k for k, v in self.tokens.items()}


    def _init_mx_dict(self) -> dict:
        my_dict = dict()
        for i in range(0, 654):
            my_dict[i] = 0
        return my_dict

    def rev_get(self, a):
        return self.rev_tokens[a]

    def get(self, a: int, token_type: str, b = False):
        if a == -1:
            b = True
            my_token = token_type
        else:
            my_token = f"{token_type}_{a}"

        if my_token in self.tokens:
            if not b:
                self.token_max[self.tokens[my_token]] += 1
            return self.tokens[my_token]
        else:
            if token_type is PITCH_TYPE:
                self.tokens[my_token] = self.pitch_position
                self.pitch_position += 1

            elif token_type is VELOCITY_TYPE:
                self.tokens[my_token] = self.velocity_position
                self.velocity_position += 1

            elif token_type is DURATION_TYPE:
                self.tokens[my_token] = self.duration_position
                self.duration_position += 1

            elif token_type is START_TYPE:
                self.tokens[my_token] = self.instruction_start_position
                self.instruction_start_position += 1

            elif token_type is SHIFT_TYPE:
                self.tokens[my_token] = self.instruction_shift_position
                self.instruction_shift_position += 1

            self.token_max[self.tokens[my_token]] = 1
            return self.tokens[my_token]

    def save(self):
        json_string = json.dumps(self.tokens)
        with open(self.save_directory + "/vocab_list.json", 'w') as file:
            file.write(json_string)

        json_s = json.dumps(self.token_max)
        with open(self.save_directory + "/vocab_max.json", 'w') as file:
            file.write(json_s)

    pass